#ifndef CX__H
#define CX__H

#include <re-pari.h>

class CX
{
  GEN _v;
  CX (GEN a, GEN b) { _v=cgetg(3,t_COMPLEX); ((GEN*)_v)[1]=a; ((GEN*)_v)[2]=b; }
  CX (GEN a)        { _v = a; }
public:
  // Construction and conversion operations
  CX (RE a = 0, RE b = 0)     { *this = CX(a.value(),b.value()); }
  CX (int a,    int b = 0)    { *this = CX(stoi(a),stoi(b)); }
  CX (long a,   long b = 0)   { *this = CX(stoi(a),stoi(b)); }
  CX (float a,  float b = 0)  { *this = CX(dbltor(a),dbltor(b)); }
  CX (double a, double b = 0) { *this = CX(dbltor(a),dbltor(b)); }
//X  inline CX (const char *a, const char *b) { x << a; y << b; }
//X  inline CX (const char *ab) { *this << ab; }
  ~CX() { }
  // Standard arithmetic operators
  CX& operator +=( const CX& a );
  CX& operator -=( const CX& a );
  CX& operator *=( const CX& a );
  CX& operator /=( const CX& a );
  friend CX operator +( const CX& a, const CX& b );
  friend CX operator -( const CX& a, const CX& b );
  friend CX operator *( const CX& a, const CX& b );
  friend CX operator /( const CX& a, const CX& b );
  friend CX operator -( const CX& a );
  CX inv  () const; // returns the inverse
  CX sqr  () const; // returns the square
  CX conj () const; // returns the conjugate
  RE norm () const; // returns the norm
  friend CX   sqr( const CX& a );
  friend CX   inv( const CX& a );
  friend CX  conj( const CX& a );
  friend RE  norm( const CX& a );
  friend CX   pow( const CX& c, const unsigned long exposant );
  friend void unit(CX& c);
  // Standard comparison operators
  friend bool operator !=( const CX& a, const CX& b );
  friend bool operator ==( const CX& a, const CX& b );
  friend long IsZero( const CX& a );
  friend long IsOne ( const CX& a );
  // Adding a product is faster with this function
  void addprod (const CX& a, const CX& b);
  // Output functions
  static FMT  output(void)  { return RE::output(); }
  friend void operator <<(char*& s, const CX& c);
  friend void operator <<(CX& c, const char* s);
};

inline ostream& operator<<(ostream& s, const CX& a)
  { char *str; str << a; s << str; free(str); return s; }

inline long IsZero ( const CX& a )  { return gcmp0(a._v); }
inline long IsOne  ( const CX& a )  { return gcmp1(a._v); }
inline RE CX::norm () const         { return RE(gnorm(_v)); }
inline CX CX::conj () const         { return CX(gconj(_v)); }
inline CX operator -( const CX& a ) { return CX(gneg(a._v)); }

inline CX& CX::operator +=( const CX& a ) { _v = gadd(_v, a._v); return *this; }
inline CX& CX::operator -=( const CX& a ) { _v = gsub(_v, a._v); return *this; }
inline CX& CX::operator *=( const CX& a ) { _v = gmul(_v, a._v); return *this; }
inline CX& CX::operator /=( const CX& a ) { _v = gdiv(_v, a._v); return *this; }

inline bool operator ==( const CX& a, const CX& b ) { return !gcmp(a._v,b._v); }
inline bool operator !=( const CX& a, const CX& b ) { return !(a==b); }

inline CX operator+(const CX& a, const CX& b) { CX z = a; z += b; return z; }
inline CX operator-(const CX& a, const CX& b) { CX z = a; z -= b; return z; }
inline CX operator*(const CX& a, const CX& b) { CX z = a; z *= b; return z; }
inline CX operator/(const CX& a, const CX& b) { CX z = a; z /= b; return z; }

inline CX CX::sqr( ) const { return CX(gsqr(_v)); }
inline CX CX::inv( ) const { return CX(ginv(_v)); }

inline void CX::addprod(const CX& a, const CX& b) { *this += a*b; }

#define DBLADD
inline CX   pow( const CX& c, const unsigned long exposant )
{
  CX res = 1;
  unsigned long k = exposant;
#ifdef DBLADD
  CX a = c;
  for( ; k > 1; k >>= 1) {
    if (k&1) res *= a;
    a *= a;
  }
  if (k) res *= a;
#else
  for( ; k > 0; k--) res *= c;
#endif
  return res;
}

#endif
