#ifndef RE__H
#define RE__H

// ********************************************************
//            REALs
// Inlude the PARI header file
extern "C" {
#include <genpari.h>
}
  extern long prec;  // precision is not in the header files, but is needed
// Make it work with pari versions older than 1.9xx
  #ifndef GSTR
  #define typeer talker
  #endif
// Redefining typ, so it is not a macro
  #undef typ
  inline ulong typ(GEN x) { return ulong(x[0]) >> TYPSHIFT; }
// Redefining gcoeff, so it can be a lvalue
  #undef gcoeff
  #define gcoeff(a,i,j) (((GEN**)a)[j][i])

class RE
{
  GEN v;
public:
  // Construction and conversion operations
  RE ()            { v = gzero; }
  RE (double a)    { v = dbltor(a); }
  //RE (float a)     { v = dbltor(a); }
  //RE (long a)      { v = stoi(a); }
  //RE (int a)       { v = stoi(a); }
  RE (const RE& a) { v = gcopy(a.v); }
  ~RE () { }
  void operator =( const RE& a ) { v = gcopy(a.v); }
  // Fundamental comparison operators
  friend bool operator ==( const RE& a, const RE& b );
  friend bool operator <=( const RE& a, const RE& b );
  friend bool IsZero( const RE& a);
  friend bool IsOne ( const RE& a);
  // Fundamental arithmetic operators
  friend RE operator -( const RE& a );
  friend RE inv  ( const RE&  a);
  friend RE sqrt ( const RE&  a);
  RE& operator +=( const RE& a );
  RE& operator *=( const RE& a );
  RE& operator -=( const RE& a );
  RE& operator /=( const RE& a );
  long log2() const; // returns the precision of a ``0''
  // Fundamental output functions
  friend void operator <<( double& z, const RE& a );
  friend void operator <<( RE& z, const char* a );
  friend void operator <<( char*& s, const RE& r );
  
  // Definitions of fast additional methods
  void operator =( const double a );
  void operator =( const int  a );
  friend RE operator +( const RE& a, const RE& b );
  friend RE operator -( const RE& a, const RE& b );
  friend RE operator *( const RE& a, const RE& b );
  friend RE operator /( const RE& a, const RE& b );
private:
  // Specific for Pari
  friend inline ulong typ( const RE& a ) { return typ(a.v); }
  friend void rprec( RE& r );
  public:
  //x GEN& value ()       { return v; }
  GEN  value () const { return v; }
};

// Fundamental comparison operators
inline bool operator ==(const RE& a, const RE& b) { return gcmp(a.v, b.v); }
inline bool operator <=(const RE& a, const RE& b) { return gcmp(a.v, b.v)<=0;}
inline bool IsZero ( const RE& a ) { return gcmp0(a.v); }
inline bool IsOne  ( const RE& a ) { return gcmp1(a.v); }
// Fundamental arithmetic operators
inline RE operator - ( const RE& a ) { RE z; z.v = gneg(a.v); return z; }
inline RE inv ( const RE& a ) { RE z; z.v = ginv(a.v); return z; }
inline RE sqrt ( const RE&  a )
  { if (t_REAL==typ(a)) { RE r; r.v = mpsqrt(a.v);      return r; }
    else                { RE r; r.v = gsqrt(a.v, prec); return r; }
  }
inline RE& RE::operator +=( const RE& a ) { gaffect(v, gadd(v, a.v)); return *this; }
inline RE& RE::operator *=( const RE& a ) { gaffect(v, gmul(v, a.v)); return *this; }
inline long RE::log2() const { return -expo(v); }
// Fundamental output functions
inline void operator <<(double& z, const RE& a)
  { z = (t_REAL==typ(a.v)) ? rtodbl(a.v) :
        (t_INT !=typ(a.v)) ? 0. :
        (3 < lgefint(a.v)) ? (double) itos(a.v) : 1024.; }
inline void operator <<(RE& r, const char* s)
  { r.v = lisexpr((char*)s);
    if (is_intreal_t(typ(r))) return;
    cerr << "Erreur de conversion : `" << s << "' n'est pas un nombre\n";
    r.v = gzero; }
// [re-io.c] void operator <<( char*& s, const RE& r );

// -------- Redefinitions of additional methods --------
inline void RE::operator =( const double a ) { v = dbltor(a); }
inline void RE::operator =( const int  a )   { v = stoi(a); }
inline RE operator+(const RE& a, const RE& b) {RE z;z.v=gadd(a.v,b.v);return z;}
inline RE operator-(const RE& a, const RE& b) {RE z;z.v=gsub(a.v,b.v);return z;}
inline RE operator*(const RE& a, const RE& b) {RE z;z.v=gmul(a.v,b.v);return z;}
inline RE operator/(const RE& a, const RE& b) {RE z;z.v=gdiv(a.v,b.v);return z;}
inline RE& RE::operator -=( const RE& a ) { v = gsub(v, a.v); return *this; }
inline RE& RE::operator /=( const RE& a ) { v = gdiv(v, a.v); return *this; }

// -------- Definitions from fundamental methods --------
// Standard arithmetic operators (as functions)
inline void add(RE& z, const RE& a, const RE& b) { z = a + b; }
inline void sub(RE& z, const RE& a, const RE& b) { z = a - b; }
inline void mul(RE& z, const RE& a, const RE& b) { z = a * b; }
// Other operators
inline RE sqr ( const RE& a ) { return a * a; }
inline bool operator !=( const RE& a, const RE& b ) { return !(a == b); }

#endif
