#ifndef RE__H
#define RE__H

// ********************************************************
//            REALs
// Inlude the PARI header file
extern "C" {
#include <genpari.h>
}
  extern long prec;  // precision is not in the header files, but is needed
// Make it work with pari versions older than 1.9xx
  #ifndef GSTR
  #define typeer talker
  #endif
// Redefining typ, so it is not a macro
  #undef typ
  inline ulong typ(GEN x) { return ulong(x[0]) >> TYPSHIFT; }
// Redefining gcoeff, so it can be a lvalue
  #undef gcoeff
  #define gcoeff(a,i,j) (((GEN**)a)[j][i])

class RE
{
  GEN g;
public:
  // Construction and conversion operations
  ~RE () { }
  RE ()            { g = gzero; }
  RE (int a)       { g = stoi(a); }
  RE (long a)      { g = stoi(a); }
  RE (float a)     { g = dbltor(a); }
  RE (double a)    { g = dbltor(a); }
  RE (const RE& a) { g = gcopy(a.g); }
  RE (const char* a);
  void operator =( const RE& a ) { g = gcopy(a.g); }
  operator double (void) const;
  operator char* (void) const;
  // Fundamental comparison operators
  friend bool operator ==( const RE& a, const RE& b );
  friend bool operator <=( const RE& a, const RE& b );
  friend bool IsZero( const RE& a);
  friend bool IsOne ( const RE& a);
  // Fundamental arithmetic operators
  friend RE operator -( const RE& a );
  friend RE inv  ( const RE&  a);
  friend RE sqrt ( const RE&  a);
  RE& operator +=( const RE& a );
  RE& operator *=( const RE& a );
  RE& operator -=( const RE& a );
  RE& operator /=( const RE& a );
  long log2() const; // returns the precision of a ``0''
  
  // Definitions of fast additional methods
  void operator =( const double a );
  void operator =( const int  a );
  friend RE operator +( const RE& a, const RE& b );
  friend RE operator -( const RE& a, const RE& b );
  friend RE operator *( const RE& a, const RE& b );
  friend RE operator /( const RE& a, const RE& b );
  friend RE sqr ( const RE& a );
private:
  // Specific for Pari
  friend inline ulong typ( const RE& a ) { return typ(a.g); }
  friend void rprec( RE& r );
  public:
  GEN genvalue () const { return g; }
  GEN* genptr () { return &g; }
};

// Construction and conversion operations
// [re-io.c] RE::RE (const char* s);
// [re-io.c] RE::operator double (void) const;
// [re-io.c] inline RE::operator char* (void) const;
// Fundamental comparison operators
inline bool operator ==(const RE& a, const RE& b) { return gcmp(a.g, b.g); }
inline bool operator <=(const RE& a, const RE& b) { return gcmp(a.g, b.g)<=0;}
inline bool IsZero ( const RE& a ) { return gcmp0(a.g); }
inline bool IsOne  ( const RE& a ) { return gcmp1(a.g); }
// Fundamental arithmetic operators
inline RE operator - ( const RE& a ) { RE z; z.g = gneg(a.g); return z; }
inline RE inv ( const RE& a ) { RE z; z.g = ginv(a.g); return z; }
inline RE sqrt ( const RE&  a )
  { if (t_REAL==typ(a)) { RE r; r.g = mpsqrt(a.g);      return r; }
    else                { RE r; r.g = gsqrt(a.g, prec); return r; }
  }
inline RE& RE::operator +=( const RE& a ) { g = gadd(g, a.g); return *this; }
inline RE& RE::operator *=( const RE& a ) { g = gmul(g, a.g); return *this; }
inline long RE::log2() const { return -expo(g); }

// -------- Redefinitions of additional methods --------
inline void RE::operator =( const double a ) { g = dbltor(a); }
inline void RE::operator =( const int  a )   { g = stoi(a); }
inline RE operator+(const RE& a, const RE& b) {RE z;z.g=gadd(a.g,b.g);return z;}
inline RE operator-(const RE& a, const RE& b) {RE z;z.g=gsub(a.g,b.g);return z;}
inline RE operator*(const RE& a, const RE& b) {RE z;z.g=gmul(a.g,b.g);return z;}
inline RE operator/(const RE& a, const RE& b) {RE z;z.g=gdiv(a.g,b.g);return z;}
inline RE& RE::operator -=( const RE& a ) { g = gsub(g, a.g); return *this; }
inline RE& RE::operator /=( const RE& a ) { g = gdiv(g, a.g); return *this; }
inline RE sqr ( const RE& a ) { RE z;z.g=gsqr(a.g);return z;}

// -------- Definitions from fundamental methods --------
inline bool operator !=( const RE& a, const RE& b ) { return !(a == b); }

#endif
